#ifndef BBB_GPIO_H
#define BBB_GPIO_H

#include "eventserver.h"
#include "threadserver.h"

/*! \file */

/*!
 * \brief GPIO direction ennumeration 
 * 
 * Specifies the direction of a given GPIO pin as either input or output.
 * Bidirectional ports are not supported.
 */
typedef enum {output=0,input=1} dir_t;

/*!
 * \brief GPIO edge sensitive ennumeration
 * 
 * Specifies the input signal transitions that will generate an interal event
 * only for GPIO pins configured as inputs. 
 */
typedef enum {edge_none=0,edge_rising,edge_falling,edge_both} edge_t;

/*! 
 * \brief GPIO ports ennumeration
 * 
 * Identifiers of the different GPIO ports available on the beaglebone black. 
 */
typedef enum {GPIO0=0,GPIO1=32,GPIO2=64,GPIO3=96} gpio_ports;

/*!
 * \brief Implmentation of a GPIO interface for ARM
 *
 * This class is a wrapper to the file system interface to GPIO signals for ARM
 * processors. The file system interface required to open and close several 
 * files to perform any action on a given GPIO pin, which this class 
 * encapsulates, providing a simple user interface.
 * 
 * This class requires some system configuration to allow standard users to 
 * access the GPIO signals. See section ?? on technical report ?? for more 
 * details on the required configuration.
 * 
 * In its simplest form, this class allows to access any available GPIO pin
 * as either an input or an ouput pin, providing functions to get the current 
 * value or set the desired value respectively.
 * 
 * More advanced features are available when the GPIO isconfigured as an input.
 * In this case, it is possible to select a sensitive transition on the input
 * signal that will generate an internal event no notify it. This event is
 * implemented as a CEvent from the iriutils library ???.
 *
 * This feature will allow an external process to wait on an input signal event
 * without continuously polling the value of the input pin.
 */
class CGPIO
{
  private:
    /**
     * \brief input pin value file descriptor
     * 
     * This attribute points to the value file of the GPIO pin when it is 
     * configured as an input, and it is used for handling the internal
     * events usign the select() function.
     */
    int input_fd;
    /**
     * \brief Identifier of the GPIO pin
     *
     * This attribute holds the unique system identifier number for the GPIO
     * pin, which is build from the port identifier and the port bit. 
     */
    int gpio_id;
    /**
     * \brief Current direction of the GPIO pin
     * 
     * This attribute holds the current direction configuration of the GPIO
     * pin. 
     */
    dir_t direction;
    /**
     * \brief Current sensitive edge for the GPIO
     * 
     * This attribute holds the kind of transition on the input signal that 
     * will generate an internal event only for GPIO pins configured as 
     * inputs.
     */
    edge_t edge;
    /* events */
    /**
     * \brief Identifier of the active transition event.
     *
     * This attribute is the string identifier for the active transition event
     * for any GPIO configured as an input. This identifier should be used by
     * any external process to wait for the desired event on the GPIO pin, and
     * therefore a copy is returned by the set_active_edge() function.
     */
    std::string edge_event_id;
    /**
     * \brief Identifier of the finish event
     * 
     * This attribute is a string identifier for the internal event used to 
     * terminate the internal thread to detect the desired transitions on the
     * input signal.
     */
    std::string finish_thread_event_id;
    /**
     * \brief Reference to the event server
     * 
     * This is a reference to the unique event server to handle the internal 
     * events.
     */
    CEventServer *event_server;
    /* threads */
    /**
     * \brief Identifier of the internal thread
     * 
     * This attribute is a string identifier for the internal thread used to 
     * monitor the desired transitions on the input value only for GPIO pins 
     * configured as inputs.
     */
    std::string edge_thread_id;
    /**
     * \brief Reference to the thread server
     * 
     * This is a reference to the unique thread server to handle the internal 
     * threads.
     */
    CThreadServer *thread_server;
  protected:
    /**
     * \brief Input transitions monitor function
     * 
     * This is the main function executed by the internal thread to detect 
     * any desired transition on the input value. This function is only called 
     * when the GPIO pin is configured as an input, and it is charge to 
     * activate the edge_event_id event whenever the desired transition is 
     * detected on the input value.
     *
     * \param param a pointer to the object itself (this) since the function is
     *              declared as static.
     *
     * \return Value passed to the process that created the thread when it 
     *         finishes execution.
     */
    static void *edge_thread(void *param);
  public:
    /*!
     * \brief Constructor
     * 
     * This constructor tries to create the file system entries for the GPIO pin
     * specified by the input parameters. If successfull, all the internal
     * attributes are initialized by default. 
     *
     * The default configuration is as an input, but without monitoring any 
     * transitions on the input value.
     *
     * A CGPIOException object may be thrown in the following cases:
     *   * Not enough privileges to access the GPIO file system
     *   * Invalid port identifier
     *
     * \param port identifier of the desired GPIO port. It must be one of the 
     *             #gpio_ports values.
     * \param pin desired bit number inside the given GPIO port. It must be a 
     *            non-negative integer between 0 and 31.
     */
    CGPIO(gpio_ports port, int pin);
    /**
     * \brief Function to change the direction of the current GPIO signal
     * 
     * This function changes the direction of the current GPIO port. If the
     * GPIO port changes from input to output, the monitoring of the input 
     * value is stopped (if it was active)
     *
     * A CGPIOException object may be thrown in the following cases:
     *   * Not enough privileges to access the GPIO file system
     *
     * \param direction The new direction identifier for the current GPIO
     *                  port. It must be one of the #dir_t values.
     */
    void set_direction(dir_t direction);
    /**
     * \brief Function to set the output value of the current GPIO signal
     * 
     * This functions changes the output value of the current GPIO signal if
     * it is configured as an output. Otherwise, this function does nothing.
     *
     * A CGPIOException object may be thrown in the following cases:
     *   * Not enough privileges to access the GPIO file system
     *
     * \param value Desired output value for the current GPIO signal. True 
     *              for a logic '1' and false for '0'.
     */
    void set_value(bool value);
    /**
     * \brief Function to get the current value of the current GPIO signal
     *
     * This function returns the current value of teh current GPIO signal if
     * it is configured as an input. Otherwise, this function does nothing.
     *
     * A CGPIOException object may be thrown in the following cases:
     *   * Not enough privileges to access the GPIO file system
     *
     * \return The current value of the input port. True for a logic '1' and
     *         false for '0'.
     */
    bool get_value(void);
    /**
     * \brief Function to enable the input monitoring
     * 
     * This function starts the monitoring process on the input value if the 
     * GPIO pin is configured as an input. If the GPIO pin is configured as 
     * an output, this function does nothing.
     * 
     * This function can also be used to stop the monitoring system by passing
     * a value of none. This will stop the internal thread and will render 
     * inactive the previously returned event identifier.
     *
     * A CGPIOException object may be thrown in the following cases:
     *   * Not enough privileges to access the GPIO file system
     *
     * \param edge identifier of the desired transition on the input value to
     *             look for. It must be one of the #edge_t value.
     *
     * \return a string with the identifier of the internal event associated to 
     *         the desired transition on the input value. This string will be
     *         empty if the GPIO is configured as an output or the desired edge
     *         is edge_none.
     */
    std::string set_active_edge(edge_t edge);
    /**
     * \brief Destructor
     * 
     * This destructor stops the monitoring process, if it is active, and free
     * all object resources.
     *
     * A CGPIOException object may be thrown in the following cases:
     *   * Not enough privileges to access the GPIO file system
     *
     */
    ~CGPIO();
};

#endif
